import { pgTable } from "drizzle-orm/pg-core";
import * as p from "drizzle-orm/pg-core";
import { createInsertSchema, createSelectSchema } from "drizzle-zod";

export const UsersTable = pgTable("users", {
	// Unique identifier on a technical aspect.
	uuid: p.uuid("uuid").unique().primaryKey().defaultRandom().notNull(),

	/*firstName: p
		.varchar("first_name", {
			length: 24,
		})
		.notNull(),

	lastName: p
		.varchar("last_name", {
			length: 24,
		})
		.notNull(),**/

	email: p
		.varchar("email", {
			length: 64,
		})
		.notNull(),

	//hash (Argonid2 hash)
	hash: p
		.varchar("hash", {
			length: 97,
		})
		.notNull(),

	//METADATA
	iat: p
		.timestamp("issued_at", {
			withTimezone: true,
		})
		.defaultNow()
		.notNull(),

	gdprAcceptance: p
		.timestamp("ack_gdpr", {
			withTimezone: true,
		})
		.defaultNow()
		.notNull(),

	isAdmin: p.boolean("is_admin").default(false).notNull(),
});
export const UsersTableSelectSchema = createSelectSchema(UsersTable);
export const UsersTableInsertSchema = createInsertSchema(UsersTable);

export const FilesTable = pgTable("files", {
	// Unique identifier on a technical aspect.
	uuid: p.uuid("uuid").unique().primaryKey().defaultRandom().notNull(),

	fileName: p
		.varchar("file_name", {
			length: 128,
		})
		.notNull(),

	checksum: p
		.varchar("checksum", {
			length: 64,
		})
		.notNull(),

	extension: p
		.varchar("extension", {
			length: 16,
		})
		.notNull(),

	groupId: p.uuid("group_id").default(null).references(() => FilesGroupTable.uuid),

	fileSize: p.integer("file_size").notNull(),

	//TODO Replace by file type reference
	fileType: p
		.uuid("file_type")
		.notNull()
		.references(() => FilesTypesTable.id),

	isRestricted: p.boolean("is_restricted").notNull(),

	isDocumentation: p.boolean("is_documentation").notNull(),

	uploadedAt: p
		.timestamp("uploaded_at", {
			withTimezone: true,
		})
		.defaultNow()
		.notNull(),

	uploadedBy: p
		.varchar("uploaded_by", {
			length: 64,
		})
		.notNull(),
});

export const FilesGroupTable = pgTable("f_groups", {
	uuid: p.uuid("uuid").unique().primaryKey().defaultRandom().notNull(),

	groupName: p
		.varchar("group_name", {
			length: 64,
		})
		.unique()
		.notNull(),
});

//TODO Files types
export const FilesTypesTable = pgTable("f_types", {
	//uuid
	id: p.uuid("id").unique().primaryKey().defaultRandom().notNull(),

	typeName: p
		.varchar("type_name", {
			length: 64,
		})
		.unique()
		.notNull(),

	mime: p
		.varchar("mime_type", {
			length: 64,
		})
		.unique()
		.notNull(),
});

export const MachinesTable = pgTable("machines", {
	id: p.uuid("id").unique().primaryKey().defaultRandom().notNull(),

	machineName: p
		.varchar("machine_name", {
			length: 64,
		})
		.notNull(),

	machineType: p
		.varchar("machine_type", {
			length: 64,
		})
		.notNull(),

	//supported files format
});

//TODO Many to Many table betwen File en Machine
export const FilesForMachinesTable = pgTable("files_for_machines", {
	id: p.uuid("id").primaryKey().notNull().defaultRandom(),

	fileId: p
		.uuid("file_id")
		.notNull()
		.references(() => FilesTable.uuid),

	machineId: p
		.uuid("machine_id")
		.notNull()
		.references(() => MachinesTable.id),
});

export const FilesTypeForMachine = pgTable("f_type_for_machines", {
	id: p.uuid("id").unique().primaryKey().defaultRandom().notNull(),

	machineId: p
		.uuid("machine_id")
		.notNull()
		.references(() => MachinesTable.id),
	fileTypeId: p
		.uuid("file_type_id")
		.notNull()
		.references(() => FilesTypesTable.id),
});
