import {
	Injectable,
	InternalServerErrorException,
	NotFoundException,
	StreamableFile,
} from "@nestjs/common";
import { DbService } from "apps/backend/src/app/db/db.service";
import {
	FilesForMachinesTable,
	FilesGroupTable,
	FilesTable,
	FilesTypeForMachine,
	FilesTypesTable,
	MachinesTable,
} from "apps/backend/src/app/db/schema";
import { StorageService } from "apps/backend/src/app/storage/storage.service";
import { eq, ilike } from "drizzle-orm";

@Injectable()
export class FilesService {
	constructor(
		private readonly storage: StorageService,
		private readonly database: DbService,
	) {}

	/**
	 * Retrieves a file from the database and storage by its unique identifier.
	 *
	 * @param fileId A string representing the unique identifier of the file.
	 * @return A Promise resolving to a StreamableFile if found.
	 * @throws NotFoundException If no file with the given identifier is found.
	 */
	public async get(fileId: string): Promise<StreamableFile> {
		const foundFiles = await this.database
			.use()
			.select()
			.from(FilesTable)
			.where(eq(FilesTable.uuid, fileId))
			.prepare("getFileFromId")
			.execute();

		if (foundFiles.length === 0)
			throw new NotFoundException("File not found", {
				description: `Identifier : ${fileId}`,
			});

		if (foundFiles.length > 1)
			console.log(
				"Multiples entries found in the database.\nInformation from the first row will be used.",
			);
		const file = foundFiles[0];
		console.log(file);

		const fileBuffer = await this.storage.read(
			file.checksum,
			file.extension,
			file.isDocumentation,
		);
		const fileInformation = await this.storage.generateInformation(
			fileBuffer,
			file.fileName,
			file.isDocumentation,
		);
		const fileNameWithoutSpaces = file.fileName.replace(/\s/g, "_");
		return new StreamableFile(fileBuffer, {
			disposition: `attachment; filename="${fileNameWithoutSpaces}.${fileInformation.fileType.extension.toLowerCase()}"`,
		});
	}

	//TODO DELETE FILE

	/**
	 * Searches for files in the database using the specified search field, limit, and offset.
	 *
	 * @param {number} limit - The maximum number of results to return.
	 * @param {number} offset - The number of results to skip before starting to return results.
	 * @param {string} searchField - The field used to search for matching file names.
	 *
	 * @return {Promise<object>} A promise that resolves to the search results.
	 */
	public async search(limit: number, offset: number, searchField: string) {
		try {
			return await this.database
				.use()
				.select()
				.from(FilesTable)
				.where(ilike(FilesTable.fileName, String(`%${searchField}%`)))
				.limit(limit)
				.offset(offset)
				.prepare("searchFiles")
				.execute();
		} catch (error) {
			throw new InternalServerErrorException(error);
		}
	}

	/**
	 * Saves a file and associates it with machines and an optional group in the database.
	 *
	 * @param {Buffer} file - The file data to be saved.
	 * @param {Map<string, unknown>} data - A map containing file and association metadata.
	 * @throws {NotFoundException} If a machine or group specified in the data does not exist in the database.
	 * @return {Promise<Object>} The inserted file record.
	 */
	public async save(file: Buffer, data: Map<string, unknown>) {
		const _machineIds = data.get("machineId").toString().split(",");

		const machinesIds = new Set<string>();
		for (const machineId of _machineIds) {
			console.log(
				`Checking if machine with ID ${machineId} exist in the database...`,
			);
			const machineExists = await this.database
				.use()
				.select({
					uuid: MachinesTable.id,
				})
				.from(MachinesTable)
				.where(eq(MachinesTable.id, machineId))
				.prepare("checkMachineExists")
				.execute();

			if (machineExists.length === 0) {
				throw new NotFoundException(`Machine with ID "${machineId}" not found`);
			}

			machinesIds.add(machineExists[0].uuid);
		}

		const _group = data.get("groupId") as string | null;
		console.log("Linking to group :\n", _group);
		/*if (!_group) {
			throw new NotFoundException(`Group with ID "${_group}" not found`);
		}*/

		// verify that the group exist in the database
		const groupExists = await this.database
			.use()
			.select()
			.from(FilesGroupTable)
			.where(eq(FilesGroupTable.uuid, _group))
			.prepare("checkGroupExists")
			.execute();

		if (_group && groupExists.length === 0) {
			throw new NotFoundException(`Group with ID "${_group}" not found`);
		}

		const saveResult = await this.storage.new(
			data.get("fileName") as string,
			file,
			_machineIds,
			Boolean(data.get("isDocumentation")),
		);
		console.log(saveResult);
		const mimeId = await this.database
			.use()
			.select()
			.from(FilesTypesTable)
			.where(eq(FilesTypesTable.mime, saveResult.fileType.mime));

		const inserted = await this.database
			.use()
			.insert(FilesTable)
			.values({
				fileName: data.get("fileName") as string,
				checksum: saveResult.fileChecksum,
				extension: saveResult.fileType.extension,
				fileSize: saveResult.fileSize,
				fileType: mimeId[0].id,
				isRestricted: Boolean(data.get("isRestricted")),
				isDocumentation: Boolean(data.get("isDocumentation")),
				uploadedBy: data.get("uploadedBy") as string,
			})
			.returning();
		if (groupExists[0].uuid) {
			await this.database.use()
				.update(FilesTable)
				// @ts-ignore TODO FIX
				.set({groupId: groupExists[0].uuid})
				.where(eq(FilesTable.uuid, inserted[0].uuid))
				.prepare("addGroupToFile")
				.execute()
		}

		console.log(inserted);

		for (const machineId of machinesIds) {
			console.log(
				`Append file ${inserted[0].fileName} for machine : "${machineId}"`,
			);
			const linkRow = await this.database
				.use()
				.insert(FilesForMachinesTable)
				.values({
					fileId: inserted[0].uuid,
					machineId: machineId,
				});
		}
		return inserted[0];
	}
}
