import * as console from "node:console";
import * as crypto from "node:crypto";
import { readFile, writeFile } from "node:fs/promises";
import { join } from "node:path";
import {
	BadRequestException,
	Injectable,
	InternalServerErrorException,
	NotFoundException,
	StreamableFile,
} from "@nestjs/common";
import { DbService } from "apps/backend/src/app/db/db.service";
import {
	FilesTypeForMachine,
	FilesTypesTable,
} from "apps/backend/src/app/db/schema";
import { IFileInformation } from "apps/backend/src/app/storage/storage.types";
import { eq } from "drizzle-orm";
import { filetypeinfo } from "magic-bytes.js";

@Injectable()
export class StorageService {
	//TODO make file size configurable by admin
	private maxFileSize = 256; // MiB unit
	constructor(private readonly dbService: DbService) {}

	/**
	 * Save a file to the specified directory.
	 *
	 * @param {string} fileName - The name of the file.
	 * @param {Buffer} file - The file content in Buffer format.
	 * @throws {InternalServerErrorException} If the file save fails.
	 * @return {Promise<void>} A Promise that resolves when the file is successfully saved.
	 */
	private async saveFile(fileName: string, file: Buffer): Promise<void> {
		try {
			await writeFile(join(process.cwd(), "assets/", fileName), file, "utf8");
			console.log(`File "${fileName}" saved.`);
		} catch (err) {
			console.error(err);
			throw new InternalServerErrorException("File save failed !");
		}
	}

	/**
	 * Checks if the current MIME type and file size meet the specified conditions.
	 * @param {Array<string>} machineIds - The IDs of the associated machines.
	 * @param {Buffer} file - The file to check.
	 * @return {Promise<boolean>} - A Promise that resolves to true if the conditions are met, false otherwise.
	 */
	private async checkConditions(
		machineIds: Array<string>,
		file: Buffer,
	): Promise<boolean> {
		/**
		 * Checks if the given MIME type is present in all machines' MIME type sets.
		 *
		 * @param {Map<string, Set<string>>} mimesForMachines - A map where the key is the machine identifier and the value is a set of MIME types supported by that machine.
		 * @param {string} currentMime - The MIME type to check for presence in all sets.
		 * @return {boolean} Returns true if the MIME type is found in all sets, otherwise false.
		 */
		function checkMime(mimesForMachines: Map<string, Set<string>>, currentMime: string): boolean {
			let notFoundCount = 0;
			for (const mimesForMachine of mimesForMachines) {
				const [key, set] = mimesForMachine
				if (!set.has(currentMime)) {notFoundCount++}
			}
			return notFoundCount === 0;
		}

		const fileType = filetypeinfo(file);

		// Array of MIMEs with possible duplicate field
		const _mimes: Array<string> = [];

		const machinesMap: Map<string, Set<string>> = new Map<string, Set<string>>()
		// Fetching MIMEs for the associated machines
		for (const machineId of machineIds) {
			console.debug(`Fetching mimeTypes for machine : ${machineId}`);
			// Get MIMEs associated to a machine
			const allowedMimeId = this.dbService
				.use()
				.select()
				.from(FilesTypeForMachine)
				.where(eq(FilesTypeForMachine.machineId, machineId))
				.as("allowedMimeId");
			const _allowedMime = await this.dbService
				.use()
				.select({
					slug: FilesTypesTable.mime,
					name: FilesTypesTable.typeName,
				})
				.from(FilesTypesTable)
				.leftJoin(
					allowedMimeId,
					eq(FilesTypesTable.id, allowedMimeId.fileTypeId),
				);
			console.debug(`Total : ${_allowedMime.length}`);
			// Append each MIME of a machine
			const tempSet = new Set<string>()
			for (const allowedMimeElement of _allowedMime) {
				tempSet.add(allowedMimeElement.slug)
			}
			machinesMap.set(machineId, tempSet)
			tempSet.clear()
		}
		//Store the MIMEs without duplicate
		const mimeSet = new Set(_mimes);
		console.debug(`Indexed ${mimeSet.size} unique mimeTypes`);

		//check file size is less than 2mb
		const fileSize = file.byteLength;
		if (fileSize > this.maxFileSize * (1024 * 1024)) {
			throw new BadRequestException("File size to high.", {
				cause: "File size",
				description: `File size exceeds the limit. Maximum file size allowed is ${this.maxFileSize}MiB.`,
			});
		}

		if (!checkMime(machinesMap, fileType[0].mime))
			throw new BadRequestException({
				cause: "MIME type",
				description: `Invalid MIME type. Allowed MIME types are: ${[...mimeSet].join(", ")}.`,
			});

		return true;
	}

	/**
	 * Retrieves the contents of a file as a Buffer.
	 * @param {string} fileName - The name of the file to retrieve.
	 * @throws {NotFoundException} If the file is not found.
	 * @returns {Promise<Buffer>} A Promise that resolves to the file contents as a Buffer.
	 */
	private async getFile(fileName: string): Promise<Buffer> {
		try {
			return await readFile(join(process.cwd(), "assets/", fileName));
		} catch (err) {
			throw new NotFoundException("File not found");
		}
	}

	/**
	 * Calculates the SHA256 checksum for a given file.
	 *
	 * @param {Buffer} file - The file to calculate the checksum for.
	 * @return {Promise<string>} A Promise resolving to the calculated checksum as a lowercase hexadecimal string.
	 * @throws {InternalServerErrorException} If an error occurs during the checksum calculation.
	 */
	public async getChecksum(file: Buffer): Promise<string> {
		return new Promise((resolve) => {
			try {
				const checksum = crypto
					.createHash("sha256")
					.update(file)
					.digest("hex")
					.toLowerCase();
				resolve(checksum);
			} catch (err) {
				throw new InternalServerErrorException(err);
			}
		});
	}

	/**
	 * Generate file information based on the provided file and display name.
	 *
	 * @param {Buffer} file - The file in Buffer format.
	 * @param {string} fileDisplayName - The display name of the file.
	 * @param {boolean} [isDocumentation] - Optional flag to indicate if the file is a documentation file.
	 * @returns {Promise<IFileInformation>} - A Promise that resolves to the generated file information.
	 */
	public async generateInformation(
		file: Buffer,
		fileDisplayName: string,
		isDocumentation?: boolean,
	): Promise<IFileInformation> {
		const fileType = filetypeinfo(file);
		const checksum = await this.getChecksum(file);
		const fileName = `${isDocumentation ? "doc" : "file"}-${checksum}.${fileType[0].extension.toLowerCase()}`;
		return {
			fileName: fileName,
			fileDisplayName: fileDisplayName,
			fileSize: file.byteLength,
			fileChecksum: checksum,
			fileType: fileType[0],
			isDocumentation: isDocumentation || false,
		};
	}

	/**
	 * Reads a file from storage based on the provided checksum, extension, and type.
	 *
	 * @param {string} checksum - The checksum of the file for validation purposes.
	 * @param {string} extension - The file extension to define the type of file.
	 * @param {boolean} isDocumentation - A flag to determine if the file is documentation or a regular file.
	 * @return {Promise<Buffer>} - A promise that resolves to the file data buffer.
	 * @throws {NotFoundException} - Throws a NotFoundException if the file is not found in storage.
	 */
	public async read(
		checksum: string,
		extension: string,
		isDocumentation: boolean,
	) {
		try {
			const fileName = `${isDocumentation ? "doc" : "file"}-${checksum}.${extension.toLowerCase()}`;
			console.log(`Fetching file "${fileName}" from storage...`);
			const file = await this.getFile(fileName);
			console.log(
				`Got a ${file.byteLength / (1024 * 1024)}MiB file from the storage.`,
			);
			return file;
		} catch (err) {
			console.log("File not found.");
			throw new NotFoundException(err);
		}
	}

	/**
	 * Processes and saves a new file with the provided information.
	 *
	 * @param {string} fileDisplayName - The display name of the file.
	 * @param {Buffer} file - The file buffer to be processed.
	 * @param {Array<string>} machinesId - An array of machine IDs associated with the file.
	 * @param {boolean} [isDocumentation] - Optional flag to indicate if the file is documentation.
	 * @return {Promise<IFileInformation>} A promise that resolves to an object containing information about the file.
	 */
	public async new(
		fileDisplayName: string,
		file: Buffer,
		machinesId: Array<string>,
		isDocumentation?: boolean,
	): Promise<IFileInformation> {
		try {
			const info = await this.generateInformation(
				file,
				fileDisplayName,
				isDocumentation,
			);
			console.log(
				`Trying to append a new file : "${info.fileDisplayName}"...\n > Checksum SHA-256 : ${info.fileChecksum}\n > Size : ${info.fileSize / (1024 * 1024)}Mio\n > File format : ${info.fileType.mime}\n`,
			);
			const condition = await this.checkConditions(machinesId, file);
			if (!condition) {
				console.warn(
					`File "${info.fileDisplayName}" did not pass the files requirement.\n${info.fileChecksum}`,
				);
			}
			console.log(
				`File "${info.fileDisplayName}" passed the files requirement successfully.\n${info.fileChecksum}`,
			);

			// Save the file with "saveFile()"
			console.log(`Saving file "${info.fileName}"...`);
			console.log(`Nom d'affichage : ${info.fileDisplayName}`);
			await this.saveFile(info.fileName, file);

			// All good we return data about the file to append it to the db.
			console.log(`File "${info.fileName}" saved successfully.`);
			return info;
		} catch (err) {
			console.error(err);
			throw err;
		}
	}
}
